/***********************************************************************
 *                                                                     *
 *   The JIAJIA Software Distributed Shared Memory System              *
 *                                                                     *
 *   Copyright (C) 1997 the Center of High Performance Computing       *
 *   of Institute of Computing Technology, Chinese Academy of          *
 *   Sciences.  All rights reserved.                                   *
 *                                                                     *
 *   Permission to use, copy, modify and distribute this software      *
 *   is hereby granted provided that (1) source code retains these     *
 *   copyright, permission, and disclaimer notices, and (2) redistri-  *
 *   butions including binaries reproduce the notices in supporting    *
 *   documentation, and (3) all advertising materials mentioning       *
 *   features or use of this software display the following            *
 *   acknowledgement: ``This product includes software developed by    *
 *   the Center of High Performance Computing, Institute of Computing  *
 *   Technology, Chinese Academy of Sciences."                         *
 *                                                                     *
 *   This program is distributed in the hope that it will be useful,   *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.              *
 *                                                                     *
 *   Center of High Performance Computing requests users of this       *
 *   software to return to dsm@water.chpc.ict.ac.cn any                *
 *   improvements that they make and grant CHPC redistribution rights. *
 *                                                                     *
 *          Author: Weiwu Hu, Weisong Shi, Zhimin Tang                 *
 * =================================================================== *
 *   This software is ported to SP2 by                                 *
 *                                                                     *
 *         M. Rasit Eskicioglu                                         *
 *         University of Alberta                                       *
 *         Dept. of Computing Science                                  *
 *         Edmonton, Alberta T6G 2H1 CANADA                            *
 * =================================================================== *
 **********************************************************************/

/** S-DSM Mocha Version 0.2                      by Kenji KISE **/
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h> 
#include <signal.h>
#include <errno.h> 
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <sys/resource.h>
#include <arpa/inet.h>

void host_set(int, char *, char *);
/** please edit the hostname for your cluster                  **/
/****************************************************************/
void set_hosts(int *host_max){
  int i;
  *host_max = 32;

  for(i=0; i<32; i++){
    char name[512];
    char addr[512];
    sprintf(name, "comp%02d.score", i);
    sprintf(addr, "192.168.208.%d", 200+i);
    host_set(i, name, addr);
  }

  /****************************************************
  host_set(0, "comp00.score", "192.168.208.200");
  host_set(1, "comp01.score", "192.168.208.201");
  host_set(2, "comp02.score", "192.168.208.202");
  host_set(3, "comp03.score", "192.168.208.203");
  host_set(4, "comp04.score", "192.168.208.204");
  host_set(5, "comp05.score", "192.168.208.205");
  host_set(6, "comp06.score", "192.168.208.206");
  host_set(7, "comp07.score", "192.168.208.207");
  ****************************************************/
}

/*** definition for the software version                      ***/
/****************************************************************/
#define PROJ "S-DSM"
#define NAME "Mocha"
#define VER  "Version 0.2r1 2005-05-20"

/** S-DSM parameters                                           **/
/****************************************************************/
#define Maxhosts    64         /* the number of maximum  hosts  */
#define Maxlocks    63         /*                               */
#define hidelock    Maxlocks   /* the last lock is used as hide */
#define Psize       (1024* 8)  /* page size in byte             */
#define Maxmempages (1024*128) /* number mem pages              */
#define Cachepages  (1024* 1)  /* number of cache pages         */
#define Homepages   (1024*32)  /* number of home  pages         */
#define Startaddr   0x70000000 /*                               */
#define Maxmemsize  (Maxmempages* Psize) /* memory size         */
#define Homesize    (Homepages  * Psize) /* total home  size    */
#define Cachesize   (Cachepages * Psize) /* total cache size    */
#define NOT_ALLOC   0x0fffffff /* used for not allocated pages  */

#define Intbytes     4         /* sizeof(int)                   */
#define MAX_RETRY    128       /* message send retry number     */
#define TIMEOUT      20        /* in milli-second               */
#define Msgheadsize  24        /* original size is 32 byte      */
#define Maxpacket    (1024*18) /* max packet size in byte, 32KB */
#define Maxmsgsize   (Maxpacket - Msgheadsize) /*               */
#define Maxmsgs      128       /* original value is  8          */
#define Maxqueue     128       /* send and recv queue entry     */
#define Maxwtnts     2000      /* (4096-8)/4/2                  */
#define Maxstacksize 8         /* maximum lock stack size       */
#define Linesize     300       /* default message size in byte  */
#define COMM_BUF     (Maxpacket * 4) /* send & recv buffer size */
#define OP_GET_MSK   0x7f      /* mask to obtain OP field       */

/** operation field of 7bit: from 0 to 127 value               **/
/****************************************************************/
#define OP_NULL      100 /* null: the message is not allocated  */
#define OP_EXIT      101 /* server: exit                        */
#define OP_GETP      110 /* server: get page                    */
#define OP_GETPGRANT 111 /* server: get page grant              */
#define OP_DIFF      112 /* server: diff                        */
#define OP_DIFFGRANT 113 /* server: diff grant                  */
#define OP_BARR      114 /* server: barrier                     */
#define OP_BARRGRANT 115 /* server: barrier grant               */
#define OP_ACQ       116 /* server: aquire                      */
#define OP_ACQGRANT  117 /* server: aquire grant                */
#define OP_WAIT      118 /* server: wait                        */
#define OP_WAITGRANT 119 /* server: wait grant                  */
#define OP_INV       120 /* server: invalidate                  */
#define OP_WTNT      121 /* server: write-notice                */
#define OP_REL       122 /* server: release                     */
#define OP_BCAST     123 /* server: broadcast                   */

/****************************************************************/
#define top      lockstack[stackptr]
#define stol(x)  (*((long *) (x))) /* string to long */
#define ltos(x)  ((char *) (&(x))) /* long to string */

#define BEGINCS {sigset_t set, old;\
                 sigemptyset(&set);\
                 sigaddset(&set, SIGIO);\
                 sigprocmask(SIG_BLOCK, &set, &old);\
                 sigio_msk = sigismember(&old, SIGIO);}

#define ENDCS   {if(sigio_msk==0) {sigset_t set;\
                 sigemptyset(&set);\
                 sigaddset(&set, SIGIO);\
                 sigprocmask(SIG_UNBLOCK, &set, NULL);}}

/** typedefs                                                   **/
/****************************************************************/
typedef void (* void_func_handler)();
typedef unsigned int   addr_t;
typedef unsigned char* address_t;
typedef enum {UNMAP, INV, RO, RW} pagestate_t;
/****************************************************************/
typedef struct evaluate{ /** to store evaluation result **/
  long long retry;
  long long getpage_resend;
  long long send_msg;
  long long reliable_msg;
  long long sendto;
  long long send_op[256];
  long long recv_op[256];
} eva_t;
/****************************************************************/
typedef struct Message{
  short op;           /* operation                              */
  short frompid;      /* pid from                               */
  short topid;        /* pid to                                 */
  short size;         /* data size                              */
  unsigned int scope; /* getpage etc.                           */
  unsigned int lock;  /* lock-id                                */
  unsigned int temp;  /* for broadcast                          */
  unsigned int seqno; /* sequence number                        */
  char data[Maxmsgsize];
} msg_t;
/****************************************************************/
typedef struct host_t {
  char name[Linesize]; /* hostname                              */
  char ip[Linesize];   /* IP address string                     */
  char addr[Linesize]; /* 4byte IP address                      */
  int  addrlen;        /* address length                        */
  int  homesize;       /* homesize of the host                  */
} host_t;
/****************************************************************/
typedef struct CommManager{
  int         snd_fds[Maxhosts]; /* send, FDs for each host     */
  fd_set      snd_set;           /* send, FD set                */
  int         snd_maxfd;         /* send, maximum FD            */
  unsigned    snd_seq[Maxhosts]; /* send, sequence number       */
  int         rcv_fds[Maxhosts]; /* recv, FDs for each host     */
  fd_set      rcv_set;           /* recv, FD set                */
  int         rcv_maxfd;         /* recv, maximum FD            */
  unsigned    rcv_seq[Maxhosts]; /* recv, sequence number       */
} CommManager;
/****************************************************************/
typedef struct{
  unsigned int homei;    /* index of the home-page              */
  unsigned int cachei;   /* index of the cache-page             */
  unsigned int homepid;  /* home pid of the page                */
} mpage_t;
/****************************************************************/
typedef struct{
  address_t addr;        /* address                             */
  char      rdnt;        /* read notice                         */
  char      wtnt;        /* write notice                        */
} mhome_t;
/****************************************************************/
typedef struct{
  pagestate_t state;     /* state {UNMAP, INV, RO, RW}          */
  address_t   addr;      /* address                             */
  address_t   twin;      /* pointer to the twin                 */
  char        wtnt;      /* write notice                        */
} mcache_t;
/****************************************************************/
typedef struct wtnt{
  struct wtnt *more;     /* use the linked list if full         */
  int   cnt;             /* count, the number of valid entry    */
  char* addr[Maxwtnts];  /* address of the notified write       */
  int   from[Maxwtnts];  /* from pid or from scope ??           */
} wtnt_t;
/****************************************************************/
typedef struct stacktype {
  int    lockid;         /* lock id                             */
  wtnt_t *wtntp;         /* pointer to the write notices        */
} mstack_t;
/****************************************************************/
typedef struct{
  int acqs[Maxhosts];    /*                                     */
  int acqscope[Maxhosts];/*                                     */
  int acqc;              /*                                     */
  int scope;             /*                                     */
  int myscope;           /*                                     */
  wtnt_t *wtntp;         /* pointer of write-notice             */
} mlock_t;

/** function prototype                                         **/
/****************************************************************/
void error_exit(int);
void bcastserver(msg_t*);
void getpage(address_t);
void msgserver(int);
void savediff(int);
void savewtnt(wtnt_t*, address_t, int);
void senddiffs();

/** global variables                                           **/
/****************************************************************/
int hostc;     /* the number of hosts */
int jia_pid;   /* process id          */
int Startport; /* start port          */

volatile int acqwait;
volatile int barrwait;
volatile int waitwait;
volatile int getpwait;
volatile int diffwait;
volatile int noclearlocks;
volatile int bc_level;     /* broadcact init level  */

mlock_t  locks[Maxlocks+1];
mstack_t   lockstack[Maxstacksize];
int        stackptr;
mpage_t  page[Maxmempages];
mhome_t  home[Homepages+1];
mcache_t cache[Cachepages+1];
unsigned int globaladdr;
long         jiamapfd;
msg_t       *diffmsg[Maxhosts];

host_t hosts[Maxhosts+1];
msg_t  msgarray[Maxmsgs];

short reqports[Maxhosts][Maxhosts]; /* 16-bit port No */
short repports[Maxhosts][Maxhosts]; /* 16-bit port No */
short req_hton[Maxhosts][Maxhosts]; /* 16-bit port No */
short rep_hton[Maxhosts][Maxhosts]; /* 16-bit port No */

CommManager creq, crep; /* comm_request and comm_reply          */

msg_t recvq[Maxqueue];  /* message recv queue                   */
msg_t sendq[Maxqueue];  /* message send queue                   */
int recvq_head, recvq_tail, recvq_cnt;
int sendq_head, sendq_tail, sendq_cnt;

int sigio_msk;
int nodestat[Maxhosts];
int running_node;
int reliable_msg_flag;
eva_t e;

/****************************************************************/
/** simple functions                                           **/
/****************************************************************/
void jia_exit(){
  printf("** pid-%02d retry: %5lld  getpage_resend: %5lld\n", 
         jia_pid, e.retry, e.getpage_resend);
}

int jia_startstat(){ return 0; }

void jia_config(int dest,int value){
  if(jia_pid==0)
    printf("jia_config is not supported in this version.\n");
}

void inittools(){
 printf("inittools is not supported in this version.\n");
}

void memprotect(char *addr, size_t len, int prot){
  int ret = mprotect(addr,len, prot);
  if(ret!=0) error_exit(__LINE__);
}

void memmap(caddr_t addr,size_t len,int prot){
  int flags = MAP_PRIVATE | MAP_FIXED;
  caddr_t ret = mmap(addr, len, prot, flags, jiamapfd, 0);
  if(ret!=addr) error_exit(__LINE__);
}

/****************************************************************/
long jia_current_time(){
  static long sec  = 0;
  static long usec = 0;
  struct timeval p;
  gettimeofday(&p, NULL);
  if(!sec){
    sec  = p.tv_sec;
    usec = p.tv_usec;
  }
  return (p.tv_sec - sec)*1000 + (p.tv_usec - usec)/1000;
}

/****************************************************************/
float jia_clock(){
  static long sec  = 0;
  static long usec = 0;
  struct timeval p;
  gettimeofday(&p, NULL);
  if(!sec){
    sec  = p.tv_sec;
    usec = p.tv_usec;
  }
  return (float)(p.tv_sec - sec) + (p.tv_usec - usec)/1000000.0;
}

/****************************************************************/
int homehost(addr_t adr){
  return page[(adr-Startaddr)/Psize].homepid;
}

/****************************************************************/
int homepage(addr_t adr){
  return page[(adr-Startaddr)/Psize].homei;
}

#define FD_TYPE_REQ 1
#define FD_TYPE_REP 0
/****************************************************************/
int create_fd(int i, int flag, int type){
  struct  sockaddr_in addr;
  int ret1=0, ret2=0, ret3=0;
  int fd = socket(AF_INET, SOCK_DGRAM, 0);
  int port = (type==FD_TYPE_REQ) ?
    reqports[jia_pid][i] : repports[jia_pid][i];

  if(type==FD_TYPE_REQ){
    int size = COMM_BUF;
    ret1 = setsockopt(fd, SOL_SOCKET, SO_RCVBUF, 
          (char*)&size, sizeof(size));
    ret2 = setsockopt(fd, SOL_SOCKET, SO_SNDBUF, 
          (char*)&size, sizeof(size));
  }

  addr.sin_family      = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port        = (flag)? htons(0) : htons(port);
  ret3 = bind(fd, (struct sockaddr*)&addr, sizeof(addr));

  if(fd==-1||ret1!=0||ret2!=0||ret3!=0) error_exit(__LINE__);
  return fd;
}

/** initialize the communication manager                       **/
/****************************************************************/
void init_CommManager(int type){
  int i,fd;
  CommManager *p = (type==FD_TYPE_REQ) ?  &creq : &crep;
  
  p->snd_maxfd = 0;
  p->rcv_maxfd = 0;
  FD_ZERO(&(p->snd_set));
  FD_ZERO(&(p->rcv_set));

  for(i=0; i<hostc; i++) { 
    p->snd_seq[i] = 0;
    p->rcv_seq[i] = 0;

    fd = create_fd(i, 0, type);
    p->rcv_fds[i]= fd;
    FD_SET(fd, &(p->rcv_set));
    p->rcv_maxfd = ((fd+1)>=p->rcv_maxfd) ? fd+1 : p->rcv_maxfd;

    if(type==FD_TYPE_REQ){
      if(0>fcntl(creq.rcv_fds[i], F_SETOWN, getpid()))
        error_exit(__LINE__);
      if(0>fcntl(creq.rcv_fds[i], F_SETFL, 
                 O_ASYNC|O_NONBLOCK|O_APPEND))
        error_exit(__LINE__);
    }
    
    fd = create_fd(i, 1, type);
    p->snd_fds[i] = fd;
    FD_SET(fd, &(p->snd_set));
    p->snd_maxfd = ((fd+1)>=p->snd_maxfd) ? fd+1 : p->snd_maxfd;
  }
}

/****************************************************************/
void set_sockaddr(struct sockaddr_in *p, int to, int from){
  p->sin_family = AF_INET;
  memcpy(&p->sin_addr, hosts[to].addr, hosts[to].addrlen);
}

/****************************************************************/
void sigio_handler(){
  static struct timeval ptime = {0, 0}; /* poll time */
  struct sockaddr_in  from, to; 
  int i, ret;
  int flag = 0;
  int s = sizeof(from);
  int servemsg = 0;
  fd_set fds = creq.rcv_set;

  while(select(creq.rcv_maxfd, &fds, NULL, NULL, &ptime)>0){
    for(i=0; i<hostc; i++){
      msg_t *recvqt =  &recvq[recvq_tail];
      
      if(i==jia_pid || !FD_ISSET(creq.rcv_fds[i], &fds)) continue;
      
      ret = recvfrom(creq.rcv_fds[i], recvqt, Maxpacket,
                     flag, (struct sockaddr *)&from, 
                     (socklen_t *)&s);
      e.recv_op[recvqt->op]++; /* evaluate */
      
      if(recvqt->op & 0x80){ /* if reliable_msg_flag==0 */
        //if(recvqt->seqno!=(creq.rcv_seq[i]+1)) 
        //error_exit(__LINE__);
      }
      else{
        set_sockaddr(&to, recvqt->frompid, recvqt->topid);
        to.sin_port = rep_hton[recvqt->frompid][recvqt->topid];
        
        ret = sendto(crep.snd_fds[i], (char *)&(recvqt->seqno),
                     sizeof(recvqt->seqno), flag,
                     (struct sockaddr *)&to, sizeof(to));
      }
      
      /*** receive only the correct message ***/
      if(recvqt->seqno>creq.rcv_seq[i]){
        BEGINCS;{
          creq.rcv_seq[i] = recvqt->seqno;
          recvq_tail = (recvq_tail+1) % Maxqueue;
          recvq_cnt++;
          if(recvq_cnt>=Maxqueue) error_exit(__LINE__);
          if(recvq_cnt==1) servemsg=1;
        }ENDCS;
      }
    }
  }
  
  msgserver(servemsg);
}

/****************************************************************/
void send_one_message(){
  static struct timeval ptime = {0, 0}; /* poll time, sec usec */
  struct sockaddr_in to,from;
  int i, ret, end, rep;
  int s = sizeof(from);
  msg_t *sendqh = &sendq[sendq_head];
  int size      = sendqh->size + Msgheadsize;
  int toproc    = sendqh->topid;
  int fromproc  = sendqh->frompid;

  fd_set fds;
  
  set_sockaddr(&to, toproc, fromproc);
  to.sin_port = req_hton[toproc][fromproc];
  
  if(!reliable_msg_flag) sendqh->op = sendqh->op | 0x80;
  //  if(sendqh->op==OP_GETP || sendqh->op==OP_GETPGRANT)
  //    sendqh->op = sendqh->op | 0x80;

  for(i=0; i<MAX_RETRY; i++){
    int arrived = 0;
  
    BEGINCS;{
      e.send_op[sendqh->op]++; /* evaluate */
      ret = sendto(creq.snd_fds[toproc], (char *)(sendqh),
                   size, 0, (struct sockaddr *)&to, sizeof(to));
      if(ret<0) error_exit(__LINE__);
    }ENDCS;
    e.sendto++; /* evaluate */
    if(sendqh->op & 0x80) return ;/** reliable_msg_flag==0 **/
    
    e.reliable_msg++; /* evaluate */
    end = jia_current_time() + TIMEOUT;
    
    while((jia_current_time() < end) && (!arrived)){
      FD_ZERO(&fds);
      FD_SET(crep.rcv_fds[toproc], &fds);
      BEGINCS;{
        ret = select(crep.rcv_maxfd, &fds, NULL, NULL, &ptime);
      }ENDCS;
      if(FD_ISSET(crep.rcv_fds[toproc], &fds)!=0) arrived=1;
    }
    
    if(arrived==1){
      BEGINCS;{
        ret = recvfrom(crep.rcv_fds[toproc], (char *)&rep, 
                       Intbytes, 0, (struct sockaddr *)&from, 
                       (socklen_t *)&s);
      }ENDCS;
      if((ret < 0) && (errno==EINTR)) error_exit(__LINE__);
      if((ret!=-1) && (rep==sendqh->seqno)) return;
    }
    e.retry++; /* evaluate */
  }
  error_exit(__LINE__);
}

/****************************************************************/
void send_one_message_to_same_process(){
  int servemsg;

  BEGINCS;{   /***** send part *****/
    msg_t *sendqh = &sendq[sendq_head];
    int toproc    = sendqh->topid;
    msg_t *recvqt =  &recvq[recvq_tail];

    if(recvq_cnt>=Maxqueue) error_exit(__LINE__);
    creq.rcv_seq[toproc] = sendqh->seqno;
    memcpy(recvqt, sendqh, Msgheadsize + sendqh->size); /** ?? **/
    recvq_cnt++;
    recvq_tail = (recvq_tail+1) % Maxqueue;
    servemsg = (recvq_cnt==1) ? 1 : 0;
  }ENDCS;
  
  msgserver(servemsg); /***** recv part *****/
}

/****************************************************************/
void send_msg(msg_t *msg){
  int outsendmsg;
  e.send_msg++;
  BEGINCS;{
    msg_t *sendqt = &sendq[sendq_tail];

    if(sendq_cnt>=Maxqueue) error_exit(__LINE__);
    memcpy(sendqt, msg, Msgheadsize+msg->size);
    creq.snd_seq[msg->topid]++; /* increment sequential num */
    sendqt->seqno = creq.snd_seq[msg->topid];
    sendq_cnt++;
    sendq_tail = (sendq_tail+1) % Maxqueue;
    outsendmsg = (sendq_cnt==1) ? 1 : 0;
  }ENDCS;
  
  while(outsendmsg==1){
    msg_t *sendqh = &sendq[sendq_head];
    if(sendqh->topid!=sendqh->frompid) send_one_message();
    else send_one_message_to_same_process();

    BEGINCS;{
      sendq_head = (sendq_head+1) % Maxqueue;
      sendq_cnt--;
      outsendmsg = (sendq_cnt>0) ? 1 : 0;
    }ENDCS;
  }
}

/** B_CAST = ON , bsendmsg                                     **/
/****************************************************************/
void broadcast(msg_t *msg){
  int root = jia_pid;
  msg->op   = (msg->op<<8) | OP_BCAST;
  msg->temp = ((root & 0xffff)<<16) | bc_level;
  bcastserver(msg); 
}

/****************************************************************/
void bcastserver(msg_t *msg){
  int mypid,child1,child2;
  int root  = (msg->temp >> 16) & 0xffff;
  int level = msg->temp & 0xffff;
  level--;
  
  mypid = ((jia_pid-root)>=0) ? 
    (jia_pid-root) : (jia_pid-root+hostc);
  child1 = mypid;
  child2 = mypid+(1<<level);
  
  if(level==0) msg->op = (msg->op>>8);
  msg->temp = ((root & 0xffff) <<16) | (level & 0xffff);
  msg->frompid = jia_pid;
  
  if(child2<hostc){
    msg->topid = (child2+root) % hostc;
    send_msg(msg); 
  }
  
  msg->topid = (child1+root) % hostc;
  send_msg(msg);
}

/*** new and free message                                      **/
/****************************************************************/
msg_t *newmsg(short op, short topid){
  static int index = -1;
  msg_t *msg;
  int i;
  if((op&OP_GET_MSK)!=op) error_exit(__LINE__);

  for(i=0; i<Maxmsgs; i++){
    index = (index+1) % Maxmsgs;
    msg   = &msgarray[index];
    if(msg->op==OP_NULL){
      msg->op      = op;
      msg->frompid = jia_pid;
      msg->topid   = topid;
      msg->size    = 0;
      return msg;
    }
  }
  error_exit(__LINE__);
  return NULL;
}

void freemsg(msg_t *msg){
  msg->op = OP_NULL;
}

/** append data to a message                                   **/
/****************************************************************/
void appendmsg(msg_t *msg, char *str, int n){
  if((msg->size+n)>Maxmsgsize) error_exit(__LINE__);
  memcpy(msg->data+msg->size, str, n);
  msg->size += n;
}

/****************************************************************/
void newtwin(address_t *twin){
  if(*twin==((address_t)NULL)) *twin=(address_t)valloc(Psize);
  if((*twin)==(address_t)NULL) error_exit(__LINE__);
}

/****************************************************************/
wtnt_t *newwtnt(){
  wtnt_t *wnptr = valloc(sizeof(wtnt_t));
  if(wnptr==NULL) error_exit(__LINE__);
  wnptr->more  = NULL;
  wnptr->cnt = 0;
  return(wnptr);
}

/****************************************************************/
void freewtntspace(wtnt_t *ptr){
  wtnt_t *last;
  wtnt_t *wtnt = ptr->more;

  while(wtnt!=NULL){
    last = wtnt;
    wtnt = wtnt->more;
    free(last);
  }
  ptr->more = NULL;
  ptr->cnt  = 0;
}

/****************************************************************/
long jia_alloc3(int size, int block, int starthost){
  int homepid;
  int mapsize;
  int allocsize;
  int homei,pagei,i;
  int originaddr = globaladdr;
  int prot = PROT_READ | PROT_WRITE;
  if((globaladdr+size)>Maxmemsize) error_exit(__LINE__);
  
  allocsize = ( size%Psize)==0 ?  size : ( size/Psize+1)*Psize;
  mapsize   = (block%Psize)==0 ? block : (block/Psize+1)*Psize;
  homepid = starthost % hostc;
  
  while(allocsize>0){
    if(jia_pid==homepid){
      if((hosts[homepid].homesize+mapsize)>=(Homepages*Psize)) 
        error_exit(__LINE__);
      
      memmap((caddr_t)(Startaddr+globaladdr), mapsize, prot);
      
      for(i=0; i<mapsize; i+=Psize){
        pagei = (globaladdr+i)/Psize;
        homei = (hosts[homepid].homesize+i)/Psize;
        home[homei].addr  = (address_t)(Startaddr+globaladdr+i);
        home[homei].wtnt  = 1;
        home[homei].rdnt  = 0;
        page[pagei].homei = homei;
      }
    }
    
    for(i=0; i<mapsize; i+=Psize){
      pagei = (globaladdr+i)/Psize;
      page[pagei].homepid = homepid;
    }
    
    hosts[homepid].homesize += mapsize;
    globaladdr += mapsize;
    allocsize  -= mapsize;   
    homepid = (homepid+1) % hostc;
  }
  
  return(Startaddr + originaddr);
}

/****************************************************************/
long jia_alloc(int size){
  static int starthost=-1;
  starthost = (starthost+1) % hostc;
  return(jia_alloc3(size, Psize, starthost));
}

long jia_alloc1 (int s)       { return(jia_alloc3(s, s, 0)); }
long jia_alloc2 (int s, int b){ return(jia_alloc3(s, b, 0)); }
long jia_alloc2p(int s, int p){ return(jia_alloc3(s, s, p)); }

/****************************************************************/
void flush_cache(int i){
  int ret   = munmap((caddr_t)cache[i].addr, Psize);
  int index = ((addr_t)cache[i].addr-Startaddr)/Psize;
  if(ret!=0) error_exit(__LINE__);

  page[index].cachei = NOT_ALLOC;
  
  cache[i].state = UNMAP;
  cache[i].wtnt  = 0;
  cache[i].addr  = 0;
}

/**  enum {UNMAP, INV, RO, RW} pagestate_t                     **/
/****************************************************************/
int generate_cache_entry(address_t addr){
  int seti = (random()>>8) % Cachepages; /* use a randam scheme */
  int invi = -1;
  int i, index;

  for(i=0; i<Cachepages; i++){
    if(cache[seti].state==UNMAP) break;
    if((invi==(-1))&&(cache[seti].state==INV)) invi = seti;
    seti = (seti+1) % Cachepages;
  }
  
  if((cache[seti].state!=UNMAP)&&(invi!=(-1))) seti = invi;

  index = seti;
  if(cache[index].state == RW){
    savediff(index);
    savewtnt(top.wtntp, cache[index].addr, Maxhosts);
    senddiffs();
    while(diffwait);
  }

  flush_cache(index);
  page[((addr_t)addr-Startaddr)/Psize].cachei = index;
  return(index);
}

/****************************************************************/
void sigsegv_handler(int signo, struct sigcontext_struct sigctx){
  int writefault;
  int cachei, homei;
  sigset_t set;
  address_t faultaddr; /* fault address */
  addr_t    faddr;     /* fault address */
  
  sigemptyset(&set);
  sigaddset(&set, SIGIO);
  sigprocmask(SIG_UNBLOCK, &set, NULL);
  
  faultaddr = (address_t)(sigctx.cr2/Psize*Psize);
  faddr     = (addr_t) faultaddr;
  writefault = ((int)(sigctx.err & 2) == 2); /* set 0 or 1 */
  
  if(faddr>=(Startaddr+globaladdr) || faddr<Startaddr){
    printf("** faddr %x is out of range (%x-%x)\n", faddr, 
           Startaddr, Startaddr+globaladdr);
    error_exit(__LINE__);
  }

  if(homehost(faddr)==jia_pid){
    memprotect((char*)faultaddr, Psize, PROT_READ|PROT_WRITE);
    homei = homepage(faddr);
    home[homei].wtnt = 1;
  }else{
    cachei = page[(faddr-Startaddr)/Psize].cachei;
    if(cachei!=NOT_ALLOC){ /** cache hit  **/
      memprotect((char*)faultaddr, Psize, PROT_READ|PROT_WRITE);
      if(!((writefault)&&(cache[cachei].state==RO))){
        getpage(faultaddr);
      }
    }else{ /** cache miss **/
      cachei = generate_cache_entry(faultaddr);
      memmap((char*)faultaddr, Psize, PROT_READ|PROT_WRITE);
      getpage(faultaddr);
    }
    
    if(writefault){
      cache[cachei].addr  = faultaddr;
      cache[cachei].state = RW;
      cache[cachei].wtnt  = 1;
      newtwin(&(cache[cachei].twin)); /*** generate a twin ***/
      memcpy(cache[cachei].twin, faultaddr, Psize);
    }else{
      cache[cachei].addr  = faultaddr;
      cache[cachei].state = RO;
      memprotect((char*)faultaddr, Psize, PROT_READ);
    }
  }
}

/** getpage, server, getprantserver                            **/
/****************************************************************/
void getpserver(msg_t *request){
  msg_t *msg;
  address_t paddr = (address_t)request->scope;
  addr_t    adr   = (addr_t)paddr;
  int       homei = homepage(adr);
  
  if(homehost(adr)!=jia_pid) error_exit(__LINE__);
  
  /******  if the page is exclusive in home ******/
  if(home[homei].rdnt==0){
    home[homei].rdnt = 1;
    home[homei].wtnt = 0;
    memprotect((caddr_t)home[homei].addr, Psize, PROT_READ);
  }
  
  msg = newmsg(OP_GETPGRANT, request->frompid);
  msg->scope = adr;

  appendmsg(msg, (char*)paddr, Psize); /* copy the page */
  send_msg(msg);
  freemsg(msg);
}

/****************************************************************/
void getpgrantserver(msg_t *msg){
  memcpy((address_t)msg->scope, msg->data, Psize);
  getpwait=0;
}

#define GETPAGE_TIMEOUT   50  /* in milli-second                */
#define GETPAGE_MAXRETRY  10  /* message send retry number      */
/****************************************************************/
void getpage(address_t addr){ /** get page **/
  int i;
  addr_t adr = (addr_t)addr;
  int homeid = homehost(adr);
  msg_t *msg = newmsg(OP_GETP, homeid);
  msg->scope = adr;

  if(homeid==jia_pid) error_exit(__LINE__);
  
  getpwait=1;
  for(i=0; i<GETPAGE_MAXRETRY; i++){
    int end = jia_current_time() + GETPAGE_TIMEOUT;
    send_msg(msg);

    while((jia_current_time()<end) && getpwait); /* busy wait */
    if(getpwait==0){
      freemsg(msg);
      return;
    }
    e.getpage_resend++;
  }
  error_exit(__LINE__);
}

/*** decode diff                                               **/
/****************************************************************/
void diffserver(msg_t *req){
  int homei;
  int stop, doffset, dsize;
  int paddr;
  msg_t *msg;
  int i = 0;
  char *data = req->data;
  int  *dint = (int *)req->data;
  
  while(i < req->size){  /***** decode_diff *****/
    paddr = dint[i>>2];
    i += Intbytes;
    stop = dint[i>>2] + i - Intbytes;
    i += Intbytes;

    homei = homepage(paddr);
    if(home[homei].wtnt!=1)
      memprotect((caddr_t)paddr, Psize, PROT_READ|PROT_WRITE);
    
    while(i<stop){
      dsize   = (dint[i>>2]    ) & 0xffff;
      doffset = (dint[i>>2]>>16) & 0xffff;
      i += Intbytes;
      memcpy((address_t)(paddr+doffset), data+i, dsize);
      i += dsize;
    }
    
    if(home[homei].wtnt!=1)
      memprotect((caddr_t)paddr, (size_t)Psize, PROT_READ);
  }
  
  msg = newmsg(OP_DIFFGRANT, req->frompid);
  send_msg(msg);
  freemsg(msg);
}

/** diffsize = 4 byte                                          **/
/****************************************************************/
int encodediff(int cachei, char* diff){
  int size  = 8; /** first 8 byte is assinged later **/
  int i     = 0;
  mcache_t *ca = &cache[cachei];
  int *twin = (int *)ca->twin;
  int *page = (int *)ca->addr;
  int pagebyte = Psize >> 2;

  for(i=0; i<pagebyte; i++){
    if(page[i]!=twin[i]){
      int header;
      int start = i<<2;
      int cnt   = 0;
      for(; i<pagebyte && page[i]!=twin[i]; i++) cnt+=Intbytes;

      header = ((start & 0xffff)<<16) | (cnt & 0xffff);
      memcpy(diff+size, &header, Intbytes);
      size += Intbytes;

      memcpy(diff+size, ca->addr+start, cnt);
      size += cnt;
    }
  }

  memcpy(diff,   &ca->addr, Intbytes);
  memcpy(diff+4, &size,     Intbytes);
  return(size);
}

/****************************************************************/
void savediff(int cachei){
  char diff[Maxmsgsize];
  int diffsize;
  int i = homehost((addr_t)cache[cachei].addr);;
  
  if(diffmsg[i]==NULL) diffmsg[i]=newmsg(OP_DIFF, i);

  diffsize = encodediff(cachei, diff);
  if((diffmsg[i]->size+diffsize)>Maxmsgsize){
    diffwait++;
    send_msg(diffmsg[i]);

    diffmsg[i]->size=0;
    appendmsg(diffmsg[i], diff, diffsize);
    while(diffwait);
  }
  else appendmsg(diffmsg[i], diff, diffsize);
}

/****************************************************************/
void senddiffs(){
  int i;
  for(i=0; i<hostc; i++)
    if(diffmsg[i]!=NULL && diffmsg[i]->size>0) diffwait++;

  for(i=0; i<hostc; i++)
    if(diffmsg[i]!=NULL && diffmsg[i]->size>0){
      send_msg(diffmsg[i]);
      freemsg(diffmsg[i]);
      diffmsg[i] = NULL;
    }
}

/****************************************************************/
void diffgrantserver(msg_t *rep){
  if(rep->size!=0) error_exit(__LINE__);
  diffwait--;
}

/****************************************************************/
void clearlocks(){
  int i;
  for(i=jia_pid; i<Maxlocks; i+=hostc){
    freewtntspace(locks[i].wtntp);
  }
}

/****************************************************************/
void endinterval(int op){
  int i;
  int hpages = hosts[jia_pid].homesize/Psize;
  
  for(i=0; i<Cachepages; i++){
    if(cache[i].wtnt==1) savediff(i);
  }
  senddiffs(); /** send diff messages **/
  
  for(i=0; i<Cachepages; i++){
    if(cache[i].wtnt==1)
      savewtnt(top.wtntp, cache[i].addr, Maxhosts);
  }

  for(i=0; i<hpages; i++){
    if(home[i].wtnt==1 && home[i].rdnt==1){ /* non-exclusive */
      savewtnt(top.wtntp, home[i].addr, Maxhosts);
      if(op==OP_BARR) home[i].rdnt=0;
    }
  }
  while(diffwait);
}

/****************************************************************/
void startinterval(){
  int i;
  int hpages = hosts[jia_pid].homesize/Psize;
  
  for(i=0; i<Cachepages; i++){
    if(cache[i].wtnt==1){
      memprotect((caddr_t)cache[i].addr, Psize, PROT_READ);
      cache[i].wtnt  = 0;
      cache[i].state = RO;
    }
  }
  
  for(i=0; i<hpages; i++){
    if(home[i].wtnt==1 && home[i].rdnt==1){  /* non-exclusive */
      home[i].wtnt=0;
      memprotect((caddr_t)home[i].addr, Psize, PROT_READ);
    }
  }
}

/****************************************************************/
void pushstack(int lock){
  stackptr++;
  if(stackptr>=Maxstacksize) error_exit(__LINE__);
  top.lockid = lock;
}

/****************************************************************/
void popstack(){
  int i;
  wtnt_t *wtnt;
  stackptr--;
  if(stackptr<-1) error_exit(__LINE__);
  
  wtnt = lockstack[stackptr+1].wtntp;
  while(wtnt!=NULL){
    for(i=0; i<wtnt->cnt; i++)
      savewtnt(top.wtntp, (address_t)wtnt->addr[i], wtnt->from[i]);
    wtnt = wtnt->more;
  }
  
  freewtntspace(lockstack[stackptr+1].wtntp);
}

/****************************************************************/
void send_wtnts(int op){
  wtnt_t *wp = top.wtntp;
  msg_t *msg = newmsg(0, (top.lockid%hostc));
  
  if(op!=OP_BARR && op!=OP_REL) error_exit(__LINE__);

  msg->scope = (op==OP_REL) ? 
    locks[hidelock].myscope : locks[top.lockid].myscope;
  msg->lock  = top.lockid;

  while(wp!=NULL){
    msg->op   = OP_WTNT;
    msg->size = 0;
    appendmsg(msg, (char *)wp->addr, wp->cnt*Intbytes);
    wp = wp->more;
    if(msg->size) send_msg(msg);
  }

  msg->op   = op; /* OP_BARR or OP_REL */
  msg->size = 0;
  send_msg(msg);
  freemsg(msg);
}

/** should blush up the implementation !!                      **/
/****************************************************************/
void savewtnt(wtnt_t *ptr, address_t addr, int frompid){
  int wtnti=0;
  int exist=0;
  wtnt_t *p = ptr;
  
  while((exist==0)&&(p!=NULL)){
    for(wtnti=0;(wtnti < p->cnt)&&
          (((unsigned long)addr/Psize)!=
           ((unsigned long)p->addr[wtnti]/Psize));
        wtnti++);
    if(wtnti >= p->cnt) p = p->more;
    else exist=1;
  }
  
  if(exist==0){
    p = ptr;
    while(p->cnt >= Maxwtnts){
      if(p->more==NULL) p->more=newwtnt();
      p = p->more;
    }
    p->addr[p->cnt] = (char *)addr;
    p->from[p->cnt] = frompid;
    p->cnt++;
  }else{
    if(ptr==locks[hidelock].wtntp){     /* barrier */
      p->from[wtnti] = Maxhosts;
    }else{
      p->from[wtnti] = frompid;     /* lock or stack */
    }
  } 
}

/****************************************************************/
void wtntserver(msg_t *msg){
  int lock = msg->lock;
  if((lock%hostc)!=jia_pid) error_exit(__LINE__);
  
  if(lock!=hidelock){       /*lock*/
    int i;
    for(i=0; i<msg->size; i+=Intbytes)
      savewtnt(locks[lock].wtntp, (address_t)stol(msg->data+i),
               locks[lock].scope);
  }else{                    /*barrier*/
    int i;
    for(i=0; i<msg->size; i+=Intbytes)
      savewtnt(locks[lock].wtntp,
               (address_t)stol(msg->data+i), msg->frompid);
  }
} 

/** invalidate cache server                                    **/
/****************************************************************/
void invserver(msg_t *msg){
  addr_t addr;
  int from;
  int lock  = msg->lock;
  int datai = 0;
  
  while(datai < msg->size){
    addr = (addr_t)stol(msg->data+datai);
    datai += Intbytes;
    
    if(lock==hidelock){      /*Barrier*/
      from = stol(msg->data + datai);
      datai += Intbytes;
    }else{                    /*Lock*/
      from = Maxhosts;
    }
    
    if((from!=jia_pid) && (homehost(addr)!=jia_pid)){
      int i = page[(addr-Startaddr)/Psize].cachei;
      if(i!=NOT_ALLOC && cache[i].state!=INV){
        cache[i].state = INV;
        cache[i].wtnt  = 0;
        memprotect((caddr_t)cache[i].addr, Psize, PROT_NONE);
      }
    }
  }
}

/****************************************************************/
wtnt_t *append_lockwtnts(msg_t *msg, wtnt_t *wtnt, int scope){
  int i;
  while(wtnt!=NULL){
    if((msg->size+(wtnt->cnt*4)) >= Maxmsgsize) break;
    for(i=0; i<wtnt->cnt; i++){
      int index = homehost((addr_t)wtnt->addr[i]);
      if(wtnt->from[i]>scope && index!=msg->topid)
        appendmsg(msg, ltos(wtnt->addr[i]), Intbytes);
    }
    wtnt = wtnt->more;
  }
  return(wtnt);
}

/****************************************************************/
void grantlock(long lock, int toproc, int acqscope){
  wtnt_t *wp = locks[lock].wtntp;
  msg_t *msg = newmsg(0, toproc);
  msg->scope = locks[lock].scope;
  msg->lock  = lock;
  
  while(wp!=NULL){
    msg->op   = OP_INV;
    msg->size = 0;
    wp = append_lockwtnts(msg, wp, acqscope);
    if(msg->size) send_msg(msg);
  }

  msg->op   = OP_ACQGRANT; 
  msg->size = 0;
  send_msg(msg);
  freemsg(msg);
}

/****************************************************************/
void relserver(msg_t *msg){
  int acqi;
  int lock = msg->lock;
  if((lock%hostc)!=jia_pid) error_exit(__LINE__);
  if(msg->frompid!=locks[lock].acqs[0]) error_exit(__LINE__);

  if(msg->scope>locks[hidelock].myscope) noclearlocks=1;
  
  locks[lock].scope++;
  
  for(acqi=0; acqi<(locks[lock].acqc-1); acqi++){
    locks[lock].acqs[acqi]     = locks[lock].acqs[acqi+1];
    locks[lock].acqscope[acqi] = locks[lock].acqscope[acqi+1];
  }
  locks[lock].acqc--;
  
  if(locks[lock].acqc>0) 
    grantlock(lock,locks[lock].acqs[0], locks[lock].acqscope[0]);
}

/****************************************************************/
void acqserver(msg_t *msg){
  int lock = msg->lock;
  if(lock%hostc!=jia_pid) error_exit(__LINE__);
  
  locks[lock].acqs[locks[lock].acqc]     = msg->frompid;
  locks[lock].acqscope[locks[lock].acqc] = msg->scope;
  locks[lock].acqc++;
  
  if(locks[lock].acqc==1) 
    grantlock(lock, locks[lock].acqs[0], locks[lock].acqscope[0]);
}

/****************************************************************/
void acquire(int lock){
  msg_t *req = newmsg(OP_ACQ, (lock%hostc));
  req->scope = locks[lock].myscope;
  req->lock  = lock;

  send_msg(req);
  freemsg(req);

  while(acqwait);
}

/****************************************************************/
void acqgrantserver(msg_t *msg){
  int lock = msg->lock;
  locks[lock].myscope = msg->scope;
  acqwait=0;
}

/****************************************************************/
wtnt_t *append_barrwtnts(msg_t *msg, wtnt_t *wtnt){
  int i;
  while(wtnt!=NULL){
    if((msg->size+(wtnt->cnt*8)) >= Maxmsgsize) break;
    for(i=0; i<wtnt->cnt; i++){
      appendmsg(msg, ltos(wtnt->addr[i]), Intbytes);
      appendmsg(msg, ltos(wtnt->from[i]), Intbytes);
    }
    wtnt = wtnt->more;
  }
  return(wtnt);
}

/****************************************************************/
void grantbarr(int lock){
  wtnt_t *wp = locks[lock].wtntp;
  msg_t *msg = newmsg(OP_INV, jia_pid);
  msg->scope = locks[lock].scope;
  msg->lock  = lock;

  while(wp!=NULL){
    msg->op   = OP_INV;
    msg->size = 0;
    wp = append_barrwtnts(msg, wp);
    if(msg->size) broadcast(msg);
  }

  msg->op   = OP_BARRGRANT;
  msg->size = 0;
  broadcast(msg);
  freemsg(msg);
}

/****************************************************************/
void barrgrantserver(msg_t *msg){
  int lock = msg->lock;
  if(noclearlocks==0) clearlocks();
  locks[lock].myscope = msg->scope;

  barrwait = 0;
  noclearlocks = 0;
}

/****************************************************************/
void barrserver(msg_t *msg){
  int lock = msg->lock;
  if(lock!=hidelock || lock%hostc!=jia_pid) error_exit(__LINE__);
  
  locks[lock].acqc++;
  
  if(locks[lock].acqc == hostc){
    locks[lock].scope++;
    grantbarr(lock);
    locks[lock].acqc=0;
    freewtntspace(locks[lock].wtntp);
  }
}

/****************************************************************/
void jia_barrier(){
  if(hostc!=1){
    if(stackptr!=0) error_exit(__LINE__);
    
    endinterval(OP_BARR);
    
    barrwait=1;
    send_wtnts(OP_BARR);
    freewtntspace(top.wtntp);
    while(barrwait);
    
    startinterval();
  }
}

/****************************************************************/
void jia_lock(int lock){
  if(hostc!=1){
    if((lock<0) || (lock>=Maxlocks)) error_exit(__LINE__);
    
    endinterval(OP_ACQ);
    acqwait=1;
    acquire(lock);
    startinterval();
    
    pushstack(lock);
  }
}

/****************************************************************/
void jia_unlock(int lock){
  if(hostc!=1){
    if((lock<0) || (lock>=Maxlocks)) error_exit(__LINE__);
    if(lock!=top.lockid) error_exit(__LINE__);
    
    endinterval(OP_REL);
    send_wtnts(OP_REL);
    startinterval();

    popstack();
  }
}

/** jia_wait, waitserver, waitgrantserver                      **/
/****************************************************************/
void jia_wait(){
  if(hostc!=1){
    msg_t *msg = newmsg(OP_WAIT, 0);
    waitwait = 1;
    send_msg(msg);
    freemsg(msg);
    
    while(waitwait) ;
  }
}

void waitserver(msg_t *req){
  static int waitcounter=0;
  waitcounter++;
  if(waitcounter==hostc){
    msg_t *msg = newmsg(OP_WAITGRANT, 0);
    waitcounter = 0;
    broadcast(msg);
    freemsg(msg);
  }
}

void waitgrantserver(msg_t *req){
 waitwait=0;
}

/** This implementation does not work well.                    **/
/****************************************************************/
void exit_all(){ /***** kill all process *****/
  int i;
  msg_t *msg = newmsg(OP_EXIT, 0);

  for(i=0; i<hostc; i++){
    if(i!=jia_pid && nodestat[i]==0){
      msg->topid = i;
      send_msg(msg);
    }
  }

  sleep(2);
  printf("** exit_all() at pid-%d\n\n", jia_pid);
  fflush(stdout);
  exit(1);
} 

/****************************************************************/
void error_exit(int line){
  printf("** pid-%02d exit. line %4d\n", jia_pid, line);
  if(running_node>1) exit_all();
  exit(1);
}

/****************************************************************/
void sigint_handler(){
  printf("** sigint_handler, shutdown process begins.\n");
  fflush(stdout);
  exit_all();
}

/****************************************************************/
void exitserver(msg_t *req){
  printf("** pid-%02d is killed!\n", jia_pid);
  fflush(stdout);
  sleep(2);
  exit(1);
}

/****************************************************************/
void msgserver(int servemsg){
  while(servemsg){
    msg_t *recvqh =  &recvq[recvq_head];
    switch(recvqh->op & OP_GET_MSK){
    case OP_GETP     :      getpserver(recvqh); break;
    case OP_GETPGRANT: getpgrantserver(recvqh); break;
    case OP_DIFF     :      diffserver(recvqh); break;
    case OP_DIFFGRANT: diffgrantserver(recvqh); break;
    case OP_BCAST    :     bcastserver(recvqh); break;
    case OP_BARR     :      barrserver(recvqh); break;
    case OP_BARRGRANT: barrgrantserver(recvqh); break;
    case OP_INV      :       invserver(recvqh); break;
    case OP_ACQ      :       acqserver(recvqh); break;
    case OP_ACQGRANT :  acqgrantserver(recvqh); break;
    case OP_REL      :       relserver(recvqh); break;
    case OP_WTNT     :      wtntserver(recvqh); break;
    case OP_WAIT     :      waitserver(recvqh); break;
    case OP_WAITGRANT: waitgrantserver(recvqh); break;
    case OP_EXIT     :      exitserver(recvqh); break;
    default: error_exit(__LINE__);
    }

    BEGINCS;{
      recvqh->op = OP_NULL;
      recvq_head = (recvq_head+1) % Maxqueue;
      recvq_cnt--;
      servemsg = (recvq_cnt>0) ? 1 : 0;
    }ENDCS;
  }
}

/****************************************************************/
void host_set(int no, char *hostname, char *addr){
  struct hostent *he;
  if(no<0 || no>=Maxhosts){
    printf("error in host_set(), no %d\n", no);
    exit(1);
  }
  if((he = gethostbyname(hostname))==NULL){
    printf("** wrong hostname in host_set() and exit.\n");
    exit(1);
  }
  strcpy(hosts[no].name, he->h_name); 
  memcpy(hosts[no].ip,   addr, strlen(addr));
  memcpy(hosts[no].addr, he->h_addr, he->h_length);
  hosts[no].addrlen = he->h_length;
}

/****************************************************************/
void generate_process(int argc, char **argv){
  char *pwd;
  int i, j;
  char cmd[Linesize];
  
  if((pwd = getenv("PWD"))==NULL) error_exit(__LINE__);
  
  for(i=1; i<hostc; i++){
    sprintf(cmd,"rsh %s \"cd %s; ./",  hosts[i].name, pwd);
    for(j=0; j<argc; j++){
      strcat(cmd, argv[j]);
      strcat(cmd, " ");
    }
    sprintf(cmd,"%s -P%d ",cmd, Startport);
    sprintf(cmd,"%s -R%d\" &",cmd, i);
    if(system(cmd)==-1) error_exit(__LINE__);
  }
}

/* by Kenji KISE */
/****************************************************************/
void check_proc_client(){
  int netd;
  char *server = hosts[0].ip;
  struct sockaddr_in addr;
  addr.sin_family      = PF_INET;
  addr.sin_port        = htons(Startport);
  addr.sin_addr.s_addr = inet_addr(server);
  
  if((netd=socket(PF_INET, SOCK_STREAM, 0))<0) 
    error_exit(__LINE__);
  if(connect(netd, (struct sockaddr *)&addr, sizeof(addr))<0)
    error_exit(__LINE__);
  write(netd, &jia_pid, sizeof(int));
}

#define ACT_SERVER_INIT  1
#define ACT_SERVER_CHECK 2
/****************************************************************/
int check_proc_server(int act){
  int i;
  int msg;
  static int fd;
  static int netd;
  struct sockaddr_in a;

  if(act==ACT_SERVER_INIT){
    a.sin_family      = PF_INET;
    a.sin_port        = htons(Startport);
    a.sin_addr.s_addr = INADDR_ANY;
    
    if((netd=socket(PF_INET, SOCK_STREAM, 0))<0) return -1;
    if(bind(netd, (struct sockaddr *)&a, sizeof(a))<0) return -1;
    if(listen(netd, 1)<0) return -1;
    return 0;
  }
  if(act==ACT_SERVER_CHECK){
    int ret = 0;
    for(i=0; i<hostc; i++) nodestat[i]=-1;
    
    for(i=1; i<hostc; i++){
      int len;
      struct sockaddr_in from;
      len = sizeof(from);
      if((fd = accept(netd, (struct sockaddr *)&from,
                      (socklen_t *)&len))<0) error_exit(__LINE__);
      else{
        read(fd, &msg, sizeof(int));
        if(msg<0) ret=-1;
        else nodestat[msg]=0;
      }
    }
    if(ret==-1) error_exit(__LINE__);
    return ret;
  }
  error_exit(__LINE__);
  return 0;
}

#define MAX_PORT_RETRY 10
/** initialze environment variables                            **/
/** use -P_port -Q_procs -R_pid options                        **/
/****************************************************************/
void init_env(int argc, char **argv){
  int i;
  int hostc_max;
  jia_pid = 0; /* default process id */
  hostc   = 4; /* default host count */
  reliable_msg_flag = 1;
  
  for(i=0; argv[i]!=NULL; i++){
    char opt[1024];
    strcpy(opt, argv[i]);
    if(opt[0]!='-') continue;
    switch(opt[1]){
    case 'P': Startport = atol(&opt[2]); break;
    case 'Q': hostc     = atoi(&opt[2]); break;
    case 'R': jia_pid   = atol(&opt[2]); break;
    case 'S': reliable_msg_flag = 0;     break;

    }
  }

  if(jia_pid==0){
    int i;
    printf("** %s/%s %s\n", PROJ, NAME, VER);
    for(i=0; i< MAX_PORT_RETRY; i++){
      Startport = 20000 + 
        (getpid() * Maxhosts * Maxhosts * 4) % 10000 + i;
      if(check_proc_server(ACT_SERVER_INIT)==0) break;
    }
    if(i>=MAX_PORT_RETRY){ /** max retry error **/
      printf("** Error in init_process()\n");
      exit(0);
    }
  }

  set_hosts(&hostc_max);

  if(hostc>hostc_max){
    printf("** hostc %d is too large.\n", hostc);
    exit(0);
  }
  if(hostc==0 || hostc>Maxhosts){
    printf("** hostc %d is out of range?\n", hostc);
    exit(0);
  }
  
  if(jia_pid==0){
    char name[1024];
    gethostname(name, 1024);
    if(strcmp(name, hosts[0].name)){
      printf("** %s: run on %s\n", name, hosts[0].name);
      exit(1);
    }
    printf("** pagesize %dKB, home %dMB(%dMB), cache %dMB, ", 
           Psize/1024, Maxmemsize/1024/1024, Homesize/1024/1024, 
           Cachesize/1024/1024);
    printf("hosts %d\n", hostc);
    printf("** reliable_msg_flag %d\n", reliable_msg_flag);
    fflush(stdout);
  }
}

/** initialze signal handler                                   **/
/****************************************************************/
void init_signal(){
  struct sigaction act;

  /****** for message send, recv ******/
  act.sa_handler = (void_func_handler)sigio_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  if(sigaction(SIGIO, &act, NULL)) error_exit(__LINE__);

  /****** for page fault ******/
  act.sa_handler = (void_func_handler)sigsegv_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_NOMASK;
  if(sigaction(SIGSEGV, &act, NULL)) error_exit(__LINE__);

  /****** for C-d signal to kill the process ******/
  act.sa_handler = (void_func_handler)sigint_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  if(sigaction(SIGINT, &act, NULL)) error_exit(__LINE__);
}

/** initialze for message                                      **/
/****************************************************************/
void init_message(){
  int i;
  for(i=0; i<Maxmsgs; i++) msgarray[i].op=OP_NULL;
  recvq_head = 0;
  recvq_tail = 0;
  sendq_head = 0;
  sendq_tail = 0;
  recvq_cnt  = 0;
  sendq_cnt  = 0;

  /***** set broadcast height or level *****/
  for(bc_level=1; (1<<bc_level)<hostc; bc_level++);
}

/** initialze for lock                                         **/
/****************************************************************/
void init_lock(){
  int i,j;
  
  for(i=0; i<=Maxlocks; i++){
    locks[i].acqc=0;
    locks[i].scope=0;
    locks[i].myscope=-1;
    for(j=0; j<Maxhosts; j++){
      locks[i].acqs[j]=-1;
      locks[i].acqscope[j]=-1;
    }
    if((i%hostc)==jia_pid) locks[i].wtntp=newwtnt();
    else                   locks[i].wtntp=NULL;
  }
  
  for(i=0; i<Maxstacksize; i++){
    lockstack[i].lockid=0;
    lockstack[i].wtntp=newwtnt();
  }
  
  stackptr=0;
  top.lockid=hidelock;
  noclearlocks=0;
}

/****************************************************************/
void init_mem(){
  int i;
  
  for(i=0; i<Maxhosts; i++) diffmsg[i]=NULL;
  diffwait=0;
  getpwait=0;
  
  for(i=0; i<=hostc; i++) hosts[i].homesize=0; 
  
  for(i=0; i<Maxmempages; i++){
    page[i].cachei  = NOT_ALLOC;
    page[i].homei   = Homepages;
    page[i].homepid = Maxhosts;
  }
  
  for(i=0; i<=Homepages; i++){
    home[i].wtnt=0;
    home[i].rdnt=0;
    home[i].addr=(address_t)0;
  }
  
  for(i=0; i<=Cachepages; i++){
    cache[i].state = UNMAP;
    cache[i].addr  = 0;
    cache[i].twin  = NULL;
    cache[i].wtnt  = 0;
  }
  
  globaladdr=0;
  jiamapfd = open("/dev/zero", O_RDWR,0);
  srand(1);
}

/****************************************************************/
void init_comm(){
  int i,j;
  int port = Startport + 1;

  for(i=0; i<hostc; i++)
    for(j=0; j<hostc; j++){
      reqports[i][j] = port; port++;
      repports[i][j] = port; port++;

      req_hton[i][j] = htons(reqports[i][j]);
      rep_hton[i][j] = htons(repports[i][j]);
    }
  
  init_CommManager(FD_TYPE_REQ);
  init_CommManager(FD_TYPE_REP);
}

/****************************************************************/
void jia_init(int argc, char **argv){
  running_node = 1;
  memset(&e, 0, sizeof(eva_t));

  init_env(argc, argv);

  init_signal();
  init_message();
  init_lock();
  init_mem();
  init_comm();

  if(jia_pid==0){
    generate_process(argc,argv);
    check_proc_server(ACT_SERVER_CHECK);
  }
  else check_proc_client();
  running_node = hostc;
  sleep(1); /* to eliminate the initial sendto retry */
}
/****************************************************************/
